{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of a multi-dimensional interval differentiation        }
{    arithmetic which allows function evaluation with automatic differen-    }
{    tiation up to first order (i.e. gradient or Jacobian).                  }
{ Method: Overloading of operators and elementary functions for operations   }
{    of data type 'GradType' and 'GTvector'. Note that all operators and     }
{    functions are implemented using the "modified call by reference" of     }
{    PASCAL-XSC (see Language Reference for details) to avoid the very       }
{    inefficient allocation of local memory for the copies of the actual     }
{    parameters.                                                             }
{ Global types, operators, functions, and procedures:                        }
{    type       GradType, GTvector: data types of differentiation arithmetic }
{    operators  +, -, *, /        : operators of differentiation arithmetic  }
{    operator   :=                : to define 'GradType' constants           }
{    function   GradVar           : to define 'GradType' variables           }
{    functions  fValue, gradValue,                                           }
{               JacValue          : to get function and derivative values    }
{    functions  sqr, sqrt, power,                                            }
{               exp, sin, cos,... : elementary functions of diff. arithmetic }
{    procedures fEvalG, fEvalJ    : to compute function value only           }
{    procedures fgEvalG, fJEvalJ  : to compute function and first derivative }
{                                   value (gradient or Jacobian)             }
{----------------------------------------------------------------------------}
module grad_ari;

use i_ari, i_util;  { interval arithmetic, interval utility functions }

{----------------------------------------------------------------------------}
{ Global type definitions and variable                                       }
{----------------------------------------------------------------------------}
global type
  GradType      = dynamic array [*] of interval;
                  { The index range must be 0..n. The component [0]          }
                  { contains the function value, the components [1],...,[n]  }
                  { contain the gradient.                                    }
  GTvector      = global dynamic array [*] of GradType;
                  { The index range must be 1..n (i.e.: 1..n,0..n).          }

var                  { The local variable 'GradOrder' is used to select the  }
  GradOrder : 0..1;  { highest order of derivative which is computed. Its    }
                     { default value is 1, and normally the gradient or the  }
                     { Jacobian matrix are computed.                         }
{----------------------------------------------------------------------------}
{ Transfer operators and functions for constants and variables               }
{----------------------------------------------------------------------------}
global operator := (var f: GradType; u: interval);       { Generate constant }
var i : integer;                                         {-------------------}
begin
  f[0]:= u;
  for i:=1 to ub(f) do f[i]:= 0;
end;

global operator := (var f: GradType; r: real);           { Generate constant }
begin                                                    {-------------------}
  f:= intval(r);
end;

global function GradVar (x : ivector)
                           : GTvector[1..ub(x)-lb(x)+1, 0..ub(x)-lb(x)+1];
var                                                      { Generate variable }
  i, k, ubd, d : integer;                                {-------------------}
begin
  ubd:= ub(x)-lb(x)+1;   d:= 1-lb(x);
  for i:=1 to ubd do
  begin
    GradVar[i,0]:= x[i-d];  { GradVar[*,0]:=x; }
    for k:=1 to ubd do
      if i=k then GradVar[i,k]:= 1 else GradVar[i,k]:= 0;
  end;
end;

global function GradVar (v : rvector)
                           : GTvector[1..ub(v)-lb(v)+1, 0..ub(v)-lb(v)+1];
var                                                      { Generate variable }
  u : ivector[lb(v)..ub(v)];                             {-------------------}
  i : integer;
begin
  for i:=lb(v) to ub(v) do u[i]:= v[i];
  GradVar:= GradVar(u);
end;

{----------------------------------------------------------------------------}
{ Access functions for the function value, the gradient, or the Jacobian     }
{----------------------------------------------------------------------------}
global function fValue (var f: GradType) : interval; { Get function value of }
begin                                                { n-dimensional scalar- }
  fValue:= f[0];                                     { valued function       }
end;                                                 {-----------------------}

global function gradValue (var f: GradType) : ivector[1..ub(f)];
var i: integer;                                      { Get gradient value of }
begin                                                { n-dimensional scalar- }
  for i:=1 to ub(f) do gradValue[i]:= f[i];          { valued function       }
end;                                                 {-----------------------}

global function fValue (var f: GTvector) : ivector[1..ub(f,1)];
begin                                                { Get function value of }
  fValue:= f[*,0];                                   { n-dimensional vector- }
end;                                                 { valued function       }
                                                     {-----------------------}

global function JacValue (var f: GTvector) : imatrix[1..ub(f,1),1..ub(f,2)];
var i, j: integer;                                   { Get Jacobian value of }
begin                                                { n-dimensional scalar- }
  for i:=1 to ub(f,1) do                             { valued function       }
    for j:=1 to ub(f,2) do                           {-----------------------}
      JacValue[i,j]:= f[i,j];
end;

{----------------------------------------------------------------------------}
{ Monadic operators + and - for GradType operands                            }
{----------------------------------------------------------------------------}
global operator + (var u : GradType ) uplus : GradType[0..ub(u)];
begin
  uplus:= u;
end;

global operator - (var u : GradType ) umin : GradType[0..ub(u)];
var
  i : integer;
begin
  umin[0]:= -u[0];
  if (GradOrder > 0) then
    for i:=1 to ub(u) do
      umin[i]:= -u[i];
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for two GradType operands                         }
{----------------------------------------------------------------------------}
global operator + (var u,v : GradType ) add : GradType[0..ub(u)];
var
  i: integer;
begin
  add[0]:= u[0]+v[0];
  if (GradOrder > 0) then
    for i:=1 to ub(u) do add[i]:= u[i]+v[i];
end;

global operator - (var u,v : GradType ) sub : GradType[0..ub(u)];
var
  i: integer;
begin
  sub[0]:= u[0]-v[0];
  if (GradOrder > 0) then
    for i:=1 to ub(u) do sub[i]:= u[i]-v[i];
end;

global operator * (var u,v : GradType ) mul : GradType[0..ub(u)];
var
  i: integer;
begin
  mul[0]:= u[0]*v[0];
  if (GradOrder > 0) then
    for i:=1 to ub(u) do mul[i]:= v[0]*u[i]+u[0]*v[i];
end;

global operator / (var u,v: GradType ) divis : GradType[0..ub(u)];
var
  h: GradType[0..ub(u)];
  i: integer;
begin
  h[0]:= u[0]/v[0];  { Can propagate 'division by zero' error }
  if (GradOrder > 0) then
    for i:=1 to ub(u) do h[i]:= (u[i]-h[0]*v[i])/v[0];
  divis:=h;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one interval and one GradType operand         }
{----------------------------------------------------------------------------}
global operator + (var u: GradType; b: interval) add : GradType[0..ub(u)];
begin
  add:= u;  add[0]:= u[0]+b;
end;

global operator - (var u: GradType; b: interval) sub : GradType[0..ub(u)];
begin
  sub:= u;  sub[0]:= u[0]-b;
end;

global operator * (var u: GradType; b: interval) mul : GradType[0..ub(u)];
var
  i: integer;
begin
  mul[0]:= u[0]*b;
  if (GradOrder > 0) then
    for i:=1 to ub(u) do mul[i]:= b*u[i];
end;

global operator / (var u: GradType; b: interval) divis : GradType[0..ub(u)];
var
  i: integer;
begin
  divis[0]:= u[0]/b;  { Can propagate 'division by zero' error }
  if (GradOrder > 0) then
    for i:=1 to ub(u) do divis[i]:= u[i]/b;
end;

global operator + (a: interval; var v: GradType) add : GradType[0..ub(v)];
begin
  add:= v;  add[0]:= a+v[0];
end;

global operator - (a: interval; var v: GradType) sub : GradType[0..ub(v)];
var
  i: integer;
begin
  sub[0]:= a-v[0];
  if (GradOrder > 0) then
    for i:=1 to ub(v) do sub[i]:= -v[i];
end;

global operator * (a: interval; var v: GradType) mul : GradType[0..ub(v)];
var
  i: integer;
begin
  mul[0]:= a*v[0];
  if (GradOrder > 0) then
    for i:=1 to ub(v) do mul[i]:= a*v[i];
end;

global operator / (a: interval; var v: GradType) divis : GradType[0..ub(v)];
var
  h: GradType[0..ub(v)];
  i: integer;
  p: interval;
begin
  h[0]:= a/v[0];  { Can propagate 'division by zero' error }
  if (GradOrder > 0) then
  begin
    p:= -h[0]/v[0];
    for i:=1 to ub(v) do h[i]:= p*v[i];
  end;
  divis:=h;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one real and one GradType operand             }
{----------------------------------------------------------------------------}
global operator + (var u: GradType; b: real) add : GradType[0..ub(u)];
begin
  add:= u + intval(b);
end;

global operator - (var u: GradType; b: real) sub : GradType[0..ub(u)];
begin
  sub:= u - intval(b);
end;

global operator * (var u: GradType; b: real) mul : GradType[0..ub(u)];
begin
  mul:= u * intval(b);
end;

global operator / (var u: GradType; b: real) divs : GradType[0..ub(u)];
begin
  divs:= u / intval(b);  { Can propagate 'division by zero' error }
end;

global operator + (a: real; var v: GradType) add : GradType[0..ub(v)];
begin   
  add:= intval(a) + v;
end;

global operator - (a: real; var v: GradType) sub : GradType[0..ub(v)];
begin
  sub:= intval(a) - v;
end;

global operator * (a: real; var v: GradType) mul : GradType[0..ub(v)];
begin
  mul:= intval(a) * v;
end;

global operator / (a: real; var v: GradType) divs: GradType[0..ub(v)];
begin
   divs:= intval(a) / v;  { Can propagate 'division by zero' error }
end;

{----------------------------------------------------------------------------}
{ Elementary functions for GradType arguments                                }
{----------------------------------------------------------------------------}
global function sqr (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  sqr[0]:= sqr(u[0]);
  if (GradOrder > 0) then
  begin
    h:= 2*u[0];
    for i:=1 to ub(u) do sqr[i]:= h*u[i];
  end;
end;

global function power (var u: GradType; k: integer) : GradType[0..ub(u)];
var
  h : interval;
  i : integer;
begin
  if (k = 0) then
    power:= 1
  else if (k = 1) then
    power:= u
  else if (k = 2) then
    power:= sqr(u)
  else
    begin
      power[0]:= power(u[0], k);
      if (GradOrder > 0) then
      begin
        h:= k * power(u[0], k-1);
        for i:=1 to ub(u) do power[i]:= h * u[i];
      end;
    end;
end;

global function sqrt (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  h:= sqrt(u[0]);  { Can propagate domain error }
  sqrt[0]:= h;
  if (GradOrder > 0) then
  begin
    h:= 0.5/h;
    for i:=1 to ub(u) do sqrt[i]:= h*u[i];
  end;
end;

global function exp (var u: GradType) : GradType[0..ub(u)];
var
  h: interval;
  i: integer;
begin
  h:= exp(u[0]);  exp[0]:= h;
  if (GradOrder > 0) then
    for i:=1 to ub(u) do exp[i]:= h*u[i];
end;

global function ln (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  ln[0]:= ln(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/u[0];
    for i:=1 to ub(u) do ln[i]:= h*u[i];
  end;
end;

global function sin (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  sin[0]:= sin(u[0]);
  if (GradOrder > 0) then
  begin
    h:= cos(u[0]);
    for i:=1 to ub(u) do sin[i]:= h*u[i];
  end;
end;

global function cos (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  cos[0]:= cos(u[0]);
  if (GradOrder > 0) then
  begin
    h:= -sin(u[0]);
    for i:=1 to ub(u) do cos[i]:= h*u[i];
  end;
end;

global function tan (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  h:= tan(u[0]);  { Can propagate domain error }
  tan[0]:= h;
  if (GradOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h:= sqr(h)+1;                         {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    for i:=1 to ub(u) do tan[i]:= h*u[i]; { So, we use the first form.       }
  end;
end;

global function cot (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  h:= cot(u[0]);  { Can propagate domain error }
  cot[0]:= h;
  if (GradOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h:= -(sqr(h)+1);                      {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    for i:=1 to ub(u) do cot[i]:= h*u[i]; { So, we use the first form.       }
  end;
end;

global function arcsin (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arcsin[0]:= arcsin(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/sqrt(1-sqr(u[0]));
    for i:=1 to ub(u) do arcsin[i]:= h*u[i];
  end;
end;

global function arccos (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arccos[0]:= arccos(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= -1/sqrt(1-sqr(u[0]));
    for i:=1 to ub(u) do arccos[i]:= h*u[i];
  end;
end;

global function arctan (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arctan[0]:= arctan(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/(1+sqr(u[0]));
    for i:=1 to ub(u) do arctan[i]:= h*u[i];
  end;
end;

global function arccot (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arccot[0]:= arccot(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= -1/(1+sqr(u[0]));
    for i:=1 to ub(u) do arccot[i]:= h*u[i];
  end;
end;

global function sinh (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  sinh[0]:= sinh(u[0]);
  if (GradOrder > 0) then
  begin
    h:= cosh(u[0]);
    for i:=1 to ub(u) do sinh[i]:= h*u[i];
  end;
end;

global function cosh (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  cosh[0]:= cosh(u[0]);
  if (GradOrder > 0) then
  begin
    h:= sinh(u[0]);
    for i:=1 to ub(u) do cosh[i]:= h*u[i];
  end;
end;

global function tanh (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  h:= tanh(u[0]);  tanh[0]:= h;
  if (GradOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h:= 1-sqr(h);                         {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    for i:=1 to ub(u) do tanh[i]:= h*u[i];{ So, we use the first form.       }
  end;
end;

global function coth (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  h:= coth(u[0]);  { Can propagate domain error }
  coth[0]:= h;
  if (GradOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h:= 1-sqr(h);                         {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    for i:=1 to ub(u) do coth[i]:= h*u[i];{ So, we use the first form.       }
  end;
end;

global function arsinh (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arsinh[0]:= arsinh(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/sqrt(1+sqr(u[0]));
    for i:=1 to ub(u) do arsinh[i]:= h*u[i];
  end;
end;

global function arcosh (var u: GradType) : GradType[0..ub(u)];
var
  i: integer;
  h: interval;
begin
  arcosh[0]:= arcosh(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/sqrt(sqr(u[0])-1);
    for i:=1 to ub(u) do arcosh[i]:= h*u[i];
  end;
end;

global function artanh (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  artanh[0]:= artanh(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/(1-sqr(u[0]));
    for i:=1 to ub(u) do artanh[i]:= h*u[i];
  end;
end;

global function arcoth (var u: GradType) : GradType[0..ub(u)];
var
 i: integer;
 h: interval;
begin
  arcoth[0]:= arcoth(u[0]);  { Can propagate domain error }
  if (GradOrder > 0) then
  begin
    h:= 1/(1-sqr(u[0]));
    for i:=1 to ub(u) do arcoth[i]:= h*u[i];
  end;
end;

{----------------------------------------------------------------------------}
{ Predefined routines for evaluation of GradType-functions                   }
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing only the function value.                           }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'GradType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{ Description: This procedure sets 'GradOrder' to 0, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value only.        }
{----------------------------------------------------------------------------}
global procedure fEvalG (function f(y:GTvector) : GradType[lb(y,2)..ub(y,2)];
                                              x : ivector;
                              var            fx : interval);
begin
  GradOrder:= 0;  fx:= fValue(f(GradVar(x)));  GradOrder:= 1;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value and the gradient value.         }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'GradType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'gx'      : returns the gradient value 'grad f(x)'.            }
{ Description: This procedure keeps 'GradOrder' = 1, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value and the      }
{    value of the gradient.                                                  }
{----------------------------------------------------------------------------}
global procedure fgEvalG (function f(y:GTvector) : GradType[lb(y,2)..ub(y,2)];
                                               x : ivector;
                               var            fx : interval;
                               var            gx : ivector);
var
  fxG : GradType[0..ub(x)];
begin
  fxG:= f(GradVar(x));  fx:= fValue(fxG);  gx:= gradValue(fxG);
end;

{----------------------------------------------------------------------------}
{ Predefined routines for evaluation of GTvector-functions (with Jacobians)  }
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing only the function value.                           }
{ Parameters:                                                                }
{    In     : 'f'       : function of type 'GTvector'.                       }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the vector function value 'f(x)'.          }
{ Description: This procedure sets 'GradOrder' to 0, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value only.        }
{----------------------------------------------------------------------------}
global procedure fEvalJ (function f(y:GTvector) : GTvector[lb(y,1)..ub(y,1),
                                                           lb(y,2)..ub(y,2)];
                                              x : ivector;
                              var            fx : ivector);
begin
  GradOrder:= 0;  fx:= fValue(f(GradVar(x)));  GradOrder:= 1;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value and the Jacobian matrix value.  }
{ Parameters:                                                                }
{    In     : 'f'       : function of type 'GTvector'.                       }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'Jx'      : returns the Jacobian value 'Jf(x)'.                }
{ Description: This procedure keeps 'GradOrder' = 1, evaluates 'f(x)' in     }
{    differentiation arithmetic, and returns the function value and the      }
{    value of the Jacobian matrix.                                           }
{----------------------------------------------------------------------------}
global procedure fJEvalJ (function f(y:GTvector) : GTvector[lb(y,1)..ub(y,1),
                                                            lb(y,2)..ub(y,2)];
                                               x : ivector;
                               var            fx : ivector;
                               var            Jx : imatrix);
var
  fxGTv : GTvector[1..ub(x),0..ub(x)];
begin
  fxGTv:= f(GradVar(x));  fx:= fValue(fxGTv);  Jx:= JacValue(fxGTv);
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  GradOrder:= 1;
end.
