{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Determine  an optimal value 'z', an optimal basic index set 'v',  }
{    and an optimal solution vector 'x' for a linear programming problem     }
{    P = (A,b,c) given in the standard form:                                 }
{                   ( z = c^t * x = max! )                                   }
{         (LP)      (     A * x = b      )                                   }
{                   (       x >= 0       ).                                  }
{ Method: An initial basic index set is determined using the 2-phases method.}
{    Then the revised simplex method described by Syslo, Deo, and Kowalic in }
{    'Discrete Optimization Algorithms with Pascal' [p.14], Prentice Hall,   }
{    New Jersey (1983) is used.                                              }
{ Global procedure:                                                          }
{    procedure RevSimplex(...)   : determines the values z, v, x for the     }
{                                  linear programming problem P=(A,b,c)      }
{----------------------------------------------------------------------------}
module rev_simp;
use mv_ari; { Matrix/vector arithmetic }

{----------------------------------------------------------------------------}
{ Purpose: Determination of the solutions of a linear optimization problem.  }
{ Parameters:                                                                }
{   In:    'A'       : matrix of constraints                                 }
{          'b'       : right-hand side of constraints                        }
{          'c'       : objective function by its coefficients                }
{   Out:   'x'       : optimal solution vector                               }
{          'v'       : optimal basic index set                               }
{          'z'       : optimal value                                         }
{          'Err'     : error flag                                            }
{ Description: For a detailed description of the revised simplex method see  }
{   Syslo, Deo, and Kowalic.                                                 }
{----------------------------------------------------------------------------}
global procedure RevSimplex(A : rmatrix; b, c : rvector;
                            var x, v : rvector; var z : real;
                            var Err : integer                );

const eps = 1E-30;                { if (|a| < eps), it is handled as (a = 0) }
      NoError            = 0;
      WrongDimension     = 1;
      NoOptimalSolution  = 8;
      NoInitialVertex    = 9;

var
   min, s            : real;
   T                 : rmatrix[1..ub(b)+2,1..ub(c)+2];
   U                 : rmatrix[1..ub(b)+2,1..ub(b)+2];
   w, y              : rvector[1..ub(c)+2];
   ex, SolutionFound : boolean;
   i, j, k, l, p, q,
   m, n, phase       : integer;

begin

  m := ub(b); n := ub(c);
  if (n <= m) or (m < 1) or (n < 1) then                  { Check dimensions }
    Err := WrongDimension                                 {------------------}
  else                                               { Err <> WrongDimension }
    begin                                            {-----------------------}
      Err := NoError; phase := 1;                           { Initialization }
      p := m+2; q := m+2; k := m+1;                         {----------------}
      SolutionFound := false;
      c := -c;

      for i:=1 to m do                                     { Compute tableau }
        for j:=1 to n do T[i,j] := A[i,j];                 {-----------------}
      for j:=1 to n do
        begin
          T[k,j] := c[j];
          s      := 0;
          for i:= 1 to m do s := s - T[i,j];
          T[p,j] := s;
        end;
      U := id(U);

      s := 0; w := 0;                 { Determine initial basic index set v, }
      for i:=1 to m do                { initial optimal value s, and initial }
        begin                         { optimal solution x                   }
          v[i] := n - m + i; {(!)}    {--------------------------------------}
          w[i] := b[i];
          s    := s - b[i];
        end;
      w[k] := 0; w[p] := s;

      repeat

        if (w[p] >= -eps*10) and (phase = 1) then             { Only phase 1 }
          begin                                 { Solution for phase 1 found }
            phase := 2; q := m+1;               {----------------------------}
          end;

        min := 0;                                      { phase 1 and phase 2 }
        for j:= 1 to n do                { Determine index  k that minimizes }
          begin                          { U[q,*] * T[*,j] (j = 1,..,n)      }
            s := U[q,*] * T[*,j];        {-----------------------------------}
            if (s < min) then
              begin
                min := s; k := j;
              end;
          end;

        if (min > -eps) then              { vector of reduced costs vanishes }
          if (phase = 1) then             {----------------------------------}
            Err := NoInitialVertex
          else                                                   { phase = 2 }
            begin                                   { Optimal solution found }
              Solutionfound := true;                {------------------------}
              z := w[q];
            end

        else
          begin                                  { Determine candidate l for }
            for i:=1 to q do                     { exchange of indices       }
              y[i] := U[i,*] * T[*,k];           {---------------------------}
            ex := true;
            for i:= 1 to m do
              if (y[i] >= eps) then
                begin
                  s := w[i]/y[i];
                  if ex or (s < min) then
                    begin
                      min := s; l := i;
                    end;
                  ex := false;
                end;


            if ex then                                  { No candidate found }
               Err := NoOptimalSolution                 {--------------------}

            else
              begin                          { Determine new basic index set }
                v[l] := k;                   { and compute new tableau       }
                s     := 1 / y[l];           {-------------------------------}
                for j:= 1 to m do U[l,j] := U[l,j] * s;
                if (l = 1) then i := 2
                           else i := 1;
                repeat
                  s := y[i];
                  w[i] := w[i] - min * s;
                  for  j:=1  to m  do U[i,j] := U[i,j] - U[l,j] * s;
                  if (i = l-1) then i := i + 2
                               else i := i + 1;
                until (i > q);
                w[l] := min;
              end;                       { Determine new basic index set ... }

          end;                           { Determine candidate l ...         }

     until SolutionFound or (Err <> NoError);

     if SolutionFound then                 { Return optimal value z, optimal }
       begin                               { basic index set v, and optimal  }
         x := 0;                           { solution vector x               }
         for i:= 1 to m do                 {---------------------------------}
           x[trunc(v[i])] := w[i];
       end;
    end;                                         { Err <> WrongDimension ... }
end; { procedure RevSimplex }

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
